@@ -11,10 +11,9 @@
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
  * Lesser General Public License for more details.
  *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the
- * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
- * Boston, MA 02110-1301, USA.
+ * You should have received a copy of the GNU Lesser General Public License
+ * along along with this program; if not, see
+ * <https://www.gnu.org/licenses/>.
  *
  * In addition, as a special exception, the copyright holders
  * give permission to link the code of this program with
@@ -114,7 +113,7 @@ static EnchantDict *
 aspell_provider_request_dict (EnchantProvider * me _GL_UNUSED, const char *const tag)
 {
 	AspellConfig *spell_config = new_aspell_config ();
-	aspell_config_replace (spell_config, "language-tag", tag);
+	aspell_config_replace (spell_config, "master", tag);
 	aspell_config_replace (spell_config, "encoding", "utf-8");
 
 	AspellCanHaveError *spell_error = new_aspell_speller (spell_config);
