@@ -12,10 +12,9 @@
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
  * Lesser General Public License for more details.
  *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the
- * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
- * Boston, MA 02110-1301, USA.
+ * You should have received a copy of the GNU Lesser General Public License
+ * along along with this program; if not, see
+ * <https://www.gnu.org/licenses/>.
  *
  * In addition, as a special exception, the copyright holders
  * give permission to link the code of this program with
@@ -46,7 +45,6 @@
 
 #include "enchant.h"
 #include "enchant-provider.h"
-#include "debug.h"
 #include "pwl.h"
 #include "relocatable.h"
 #include "configmake.h"
@@ -62,7 +60,7 @@ struct str_enchant_broker
 	gchar * error;
 };
 
-typedef struct str_enchant_session
+typedef struct
 {
 	GHashTable *session_include;
 	GHashTable *session_exclude;
@@ -80,7 +78,7 @@ typedef struct str_enchant_session
 	EnchantProvider * provider;
 } EnchantSession;
 
-typedef struct str_enchant_dict_private_data
+typedef struct
 {
 	unsigned int reference_count;
 	EnchantSession* session;
@@ -93,7 +91,7 @@ typedef void             (*EnchantPreConfigureFunc) (EnchantProvider * provider,
 /********************************************************************************/
 
 /* Relocate a path and ensure the result is allocated on the heap */
-char *
+_GL_ATTRIBUTE_MALLOC char *
 enchant_relocate (const char *path)
 {
 	char *newpath = (char *) relocate (path);
@@ -159,20 +157,6 @@ enchant_get_conf_dirs (void)
 /********************************************************************************/
 /********************************************************************************/
 
-/* returns TRUE if tag is valid
- * for requires alphanumeric ASCII or underscore
- */
-static G_GNUC_PURE int
-enchant_is_valid_dictionary_tag(const char * const tag)
-{
-	const char * it;
-	for (it = tag; *it; ++it)
-		if(!g_ascii_isalnum(*it) && *it != '_')
-			return 0;
-
-	return it != tag; /*empty tag invalid*/
-}
-
 static char *
 enchant_normalize_dictionary_tag (const char * const dict_tag)
 {
@@ -184,26 +168,30 @@ enchant_normalize_dictionary_tag (const char * const dict_tag)
 	/* strip off en_GB.UTF-8 */
 	*strchrnul (new_tag, '.') = '\0';
 
-	/* turn en-GB into en_GB */
-	char * needle;
-	if ((needle = strchr (new_tag, '-')) != NULL)
-		*needle = '_';
-
-	/* everything before first '_' is converted to lower case */
-	needle = strchrnul (new_tag, '_');
-	for (gchar *it = new_tag; it != needle; ++it)
+	/* everything before first '_' or '-' is converted to lower case */
+	gchar *it = new_tag;
+	for (; *it != '\0' && *it != '-' && *it != '_'; ++it)
 		*it = g_ascii_tolower (*it);
-	/* everything after first '_' is converted to upper case */
-	for (gchar *it = needle; *it; ++it)
-		*it = g_ascii_toupper (*it);
+
+	/* turn en-GB into en_GB */
+	if (*it == '-')
+		*it = '_';
+
+	/* first run of alphanumberics after first '_' is converted to upper
+	   case */
+	if (*it != '\0')
+		for (++it; g_ascii_isalnum (*it); ++it)
+			*it = g_ascii_toupper (*it);
 
 	return new_tag;
 }
 
-static char *
+static _GL_ATTRIBUTE_MALLOC char *
 enchant_iso_639_from_tag (const char * const dict_tag)
 {
 	char * new_tag = strdup (dict_tag);
+	if (new_tag == NULL)
+		return NULL;
 	char * needle = strchr (new_tag, '_');
 
 	if (needle != NULL)
@@ -231,10 +219,10 @@ enchant_session_destroy (EnchantSession * session)
 
 static EnchantSession *
 enchant_session_new_with_pwl (EnchantProvider * provider,
-			      const char * const pwl,
-			      const char * const excl,
-			      const char * const lang,
-			      gboolean fail_if_no_pwl)
+				  const char * const pwl,
+				  const char * const excl,
+				  const char * const lang,
+				  gboolean fail_if_no_pwl)
 {
 	EnchantPWL *personal = NULL;
 	if (pwl)
@@ -352,12 +340,6 @@ enchant_session_clear_error (EnchantSession * session)
 /********************************************************************************/
 /********************************************************************************/
 
-static void
-enchant_free_string_list (char ** string_list)
-{
-	g_strfreev (string_list);
-}
-
 void
 enchant_dict_set_error (EnchantDict * dict, const char * const err)
 {
@@ -410,57 +392,35 @@ enchant_dict_check (EnchantDict * dict, const char *const word, ssize_t len)
 	return -1;
 }
 
-/* @suggs must have at least n_suggs + n_new_suggs space allocated
- * @n_suggs is the number if items currently appearing in @suggs
- *
- * returns the number of items in @suggs after merge is complete
- */
-static int
-enchant_dict_merge_suggestions(char ** suggs, size_t n_suggs, char ** new_suggs, size_t n_new_suggs)
-{
-	for (size_t i = 0; i < n_new_suggs; i++)
-		{
-			char * normalized_new_sugg = g_utf8_normalize (new_suggs[i], -1, G_NORMALIZE_NFD);
-
-			int is_duplicate = 0;
-			for (size_t j = 0; !is_duplicate && j < n_suggs; j++)
-				{
-					char* normalized_sugg = g_utf8_normalize (suggs[j], -1, G_NORMALIZE_NFD);
-					is_duplicate = strcmp (normalized_sugg, normalized_new_sugg) == 0;
-					g_free (normalized_sugg);
-				}
-			g_free (normalized_new_sugg);
-
-			if (!is_duplicate)
-				suggs[n_suggs++] = strdup (new_suggs[i]);
-		}
-
-	return n_suggs;
-}
-
-static char **
-enchant_dict_get_good_suggestions(EnchantDict * dict, char ** suggs, size_t n_suggs, size_t* out_n_filtered_suggs)
+/* Filter out suggestions that are invalid UTF-8 or in the exclude list. */
+static size_t
+enchant_dict_filter_suggestions(EnchantDict * dict, char ** suggs, size_t n_suggs)
 {
 	EnchantSession * session = ((EnchantDictPrivateData*)dict->enchant_private_data)->session;
 
-	char ** filtered_suggs = g_new0 (char *, n_suggs + 1);
 	size_t n_filtered_suggs = 0;
 	for (size_t i = 0; i < n_suggs; i++)
 		{
 			size_t sugg_len = strlen(suggs[i]);
 
-			if (sugg_len == 0)
-				continue;
-
 			if (g_utf8_validate(suggs[i], sugg_len, NULL) &&
-			    !enchant_session_exclude(session, suggs[i], sugg_len))
-				filtered_suggs[n_filtered_suggs++] = strdup (suggs[i]);
+				!enchant_session_exclude(session, suggs[i], sugg_len))
+				{
+					if (n_filtered_suggs != i)
+						{
+							suggs[n_filtered_suggs] = suggs[i];
+							suggs[i] = NULL;
+						}
+					n_filtered_suggs++;
+				}
+			else
+				{
+					free(suggs[i]);
+					suggs[i] = NULL;
+				}
 		}
 
-	if (out_n_filtered_suggs)
-		*out_n_filtered_suggs = n_filtered_suggs;
-
-	return filtered_suggs;
+	return n_filtered_suggs;
 }
 
 char **
@@ -475,42 +435,30 @@ enchant_dict_suggest (EnchantDict * dict, const char *const word, ssize_t len, s
 	g_return_val_if_fail (len, NULL);
 	g_return_val_if_fail (g_utf8_validate(word, len, NULL), NULL);
 
-	size_t n_dict_suggs = 0, n_suggsT = 0;
-	char **dict_suggs = NULL, **suggsT;
-
 	EnchantSession * session = ((EnchantDictPrivateData*)dict->enchant_private_data)->session;
 	enchant_session_clear_error (session);
 
 	/* Check for suggestions from provider dictionary */
-	if (dict->suggest)
+	size_t n_dict_suggs = 0;
+	char **dict_suggs = NULL;
+	if (dict->suggest != NULL)
 		{
 			dict_suggs = (*dict->suggest) (dict, word, len, &n_dict_suggs);
-			if (dict_suggs)
+			if (dict_suggs != NULL)
 				{
-					suggsT = enchant_dict_get_good_suggestions(dict, dict_suggs, n_dict_suggs, &n_suggsT);
-					enchant_free_string_list (dict_suggs);
-					dict_suggs = suggsT;
-					n_dict_suggs = n_suggsT;
+					n_dict_suggs = enchant_dict_filter_suggestions(dict, dict_suggs, n_dict_suggs);
+					if (n_dict_suggs == 0)
+						{
+							g_strfreev (dict_suggs);
+							dict_suggs = NULL;
+						}
 				}
 		}
 
-	/* Clone suggestions, if any */
-	char **suggs = NULL;
-	size_t n_suggs = n_dict_suggs;
-	if (n_suggs > 0)
-		{
-			suggs = g_new0 (char *, n_suggs + 1);
-			n_suggs = 0;
-			if (dict_suggs != NULL)
-				n_suggs = enchant_dict_merge_suggestions(suggs, n_suggs, dict_suggs, n_dict_suggs);
-		}
-
-	g_strfreev(dict_suggs);
-
 	if (out_n_suggs)
-		*out_n_suggs = n_suggs;
+		*out_n_suggs = n_dict_suggs;
 
-	return suggs;
+	return dict_suggs;
 }
 
 static void
@@ -766,7 +714,7 @@ enchant_load_providers_in_dir (EnchantBroker * broker, const char *dir_name)
 						{
 							EnchantProviderInitFunc init_func;
 							if (g_module_symbol (module, "init_enchant_provider", (gpointer *) (&init_func))
-							    && init_func)
+								&& init_func)
 								{
 									provider = init_func ();
 									if (!enchant_provider_is_valid(provider))
@@ -797,7 +745,7 @@ enchant_load_providers_in_dir (EnchantBroker * broker, const char *dir_name)
 					/* optional entry point to allow modules to look for associated files */
 					EnchantPreConfigureFunc conf_func;
 					if (g_module_symbol (module, "configure_enchant_provider", (gpointer *) (&conf_func))
-					    && conf_func)
+						&& conf_func)
 						{
 							conf_func (provider, dir_name);
 							if (!enchant_provider_is_valid(provider))
@@ -836,6 +784,7 @@ enchant_load_ordering_from_file (EnchantBroker * broker, const char * file)
 	if (!ch)
 		return;
 
+	g_debug("reading ordering file %s", file);
 	gchar *line;
 	gsize terminator;
 	while (G_IO_STATUS_NORMAL == g_io_channel_read_line (ch, &line, NULL, &terminator, NULL)) {
@@ -1054,11 +1003,13 @@ enchant_broker_request_dict_with_pwl (EnchantBroker * broker, const char *const
 	enchant_broker_clear_error (broker);
 
 	char * normalized_tag = enchant_normalize_dictionary_tag (tag);
-	if(!enchant_is_valid_dictionary_tag(normalized_tag))
-		enchant_broker_set_error (broker, "invalid tag character found");
-	else if ((dict = _enchant_broker_request_dict (broker, normalized_tag, pwl)) == NULL)
+	if ((dict = _enchant_broker_request_dict (broker, normalized_tag, pwl)) == NULL)
 		{
 			char * iso_639_only_tag = enchant_iso_639_from_tag (normalized_tag);
+			if (iso_639_only_tag == NULL) {
+				free (normalized_tag);
+				return NULL;
+			}
 			dict = _enchant_broker_request_dict (broker, iso_639_only_tag, pwl);
 			free (iso_639_only_tag);
 		}
@@ -1109,10 +1060,12 @@ enchant_broker_list_dicts (EnchantBroker * broker, EnchantDictDescribeFn fn, voi
 	GHashTable *tag_map = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
 
 	enchant_broker_clear_error (broker);
+	g_debug("enchant_broker_list_dicts");
 
 	for (GSList *list = broker->provider_list; list != NULL; list = g_slist_next (list))
 		{
 			EnchantProvider *provider = (EnchantProvider *) list->data;
+			g_debug("provider %s", provider->describe(provider));
 
 			size_t n_dicts;
 			char ** dicts = (*provider->list_dicts) (provider, &n_dicts);
@@ -1120,22 +1073,22 @@ enchant_broker_list_dicts (EnchantBroker * broker, EnchantDictDescribeFn fn, voi
 			for (size_t i = 0; i < n_dicts; i++)
 				{
 					const char * tag = dicts[i];
-					if (enchant_is_valid_dictionary_tag (tag)) {
-						GSList *providers = enchant_get_ordered_providers (broker, tag);
-						gint this_priority = g_slist_index (providers, provider);
-						if (this_priority != -1) {
-							gint min_priority = this_priority + 1;
-							gpointer ptr = g_hash_table_lookup (tag_map, tag);
-							if (ptr != NULL)
-								min_priority = g_slist_index (providers, ptr);
-							if (this_priority < min_priority)
-								g_hash_table_insert (tag_map, strdup (tag), provider);
-						}
-						g_slist_free (providers);
+					g_debug("tag %s", tag);
+					GSList *providers = enchant_get_ordered_providers (broker, tag);
+					gint this_priority = g_slist_index (providers, provider);
+					g_debug("priority %d", this_priority);
+					if (this_priority != -1) {
+						gint min_priority = this_priority + 1;
+						gpointer ptr = g_hash_table_lookup (tag_map, tag);
+						if (ptr != NULL)
+							min_priority = g_slist_index (providers, ptr);
+						if (this_priority < min_priority)
+							g_hash_table_insert (tag_map, strdup (tag), provider);
 					}
+					g_slist_free (providers);
 				}
 
-			enchant_free_string_list (dicts);
+			g_strfreev (dicts);
 		}
 
 	GSList *tags = NULL;
@@ -1199,7 +1152,7 @@ enchant_provider_dictionary_exists (EnchantProvider * provider, const char * con
 					break;
 				}
 
-			enchant_free_string_list (dicts);
+			g_strfreev (dicts);
 		}
 
 	return exists;
@@ -1234,11 +1187,13 @@ enchant_broker_dict_exists (EnchantBroker * broker, const char * const tag)
 	char * normalized_tag = enchant_normalize_dictionary_tag (tag);
 	int exists = 0;
 
-	if(!enchant_is_valid_dictionary_tag(normalized_tag))
-		enchant_broker_set_error (broker, "invalid tag character found");
-	else if ((exists = _enchant_broker_dict_exists (broker, normalized_tag)) == 0)
+	if ((exists = _enchant_broker_dict_exists (broker, normalized_tag)) == 0)
 		{
 			char * iso_639_only_tag = enchant_iso_639_from_tag (normalized_tag);
+			if (iso_639_only_tag == NULL) {
+				free (normalized_tag);
+				return 0;
+			}
 
 			if (strcmp (normalized_tag, iso_639_only_tag) != 0)
 				exists = _enchant_broker_dict_exists (broker, iso_639_only_tag);
@@ -1333,7 +1288,7 @@ enchant_broker_set_ordering (EnchantBroker * broker, const char * const tag, con
 		{
 			/* we will free ordering_dupl && tag_dupl when the hash is destroyed */
 			g_hash_table_insert (broker->provider_ordering, (gpointer)tag_dupl,
-					     (gpointer)(ordering_dupl));
+						 (gpointer)(ordering_dupl));
 		}
 	else
 		{
@@ -1388,7 +1343,7 @@ enchant_get_user_language(void)
 }
 
 
-char *
+_GL_ATTRIBUTE_MALLOC char *
 enchant_get_prefix_dir(void)
 {
 	return enchant_relocate (INSTALLPREFIX);
